import time
from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np


class FunctionAndDerivatives(ABC):
    """
    Interface for functions and their derivatives for convenient training data
    sampling. This class is meant for objects that work solely with NumPy arrays.
    The input and output dimensions are fixed (provided in the constructor) and
    used for sampling training data.

    Attributes:
        input_dimension: The dimension of the input vector.
        output_dimension: The dimension of the output vector.
        output_computation_time: Optional time to compute outputs.
        jacobian_product_computation_time: Optional time to evaluate jacobian products.
    """

    @property
    def input_dimension(self) -> int:
        """
        The dimension of the input vector.

        Returns:
            An integer representing the input dimension.
        """
        return self._input_dimension

    @property
    def output_dimension(self) -> int:
        """
        The dimension of the output vector.

        Returns:
            An integer representing the output dimension.
        """
        return self._output_dimension

    @abstractmethod
    def sample_input(self) -> np.ndarray:
        """
        Sample an input vector from the function's domain. This method should
        return a NumPy array of shape (input_dimension,).

        Returns:
            A NumPy array representing a single input sample.
        """
        ...

    def set_matrix_jacobian_prod(self, *, matrix: Optional[np.ndarray] = None) -> None:
        """
        Set the matrix used for the matrix–Jacobian product computation.

        Parameters:
            matrix: A NumPy array used to modify the Jacobian product, or None.
        """
        self._matrix_jacobian_prod_matrix = matrix

    @abstractmethod
    def value_and_matrix_jacobian_prod(self, input_sample: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the function value and the matrix–Jacobian product at a given
        input sample. The computation uses the matrix set by
        set_matrix_jacobian_prod.

        Parameters:
            input_sample: A NumPy array of shape (input_dimension,).

        Returns:
            A tuple containing:
              - A NumPy array of shape (output_dimension,) representing the
                function value.
              - A NumPy array representing the matrix–Jacobian product. Its shape
                depends on the applied matrix.
        """
        ...

    def extract_and_clear_computation_times(self) -> Dict[str, Optional[float]]:
        """
        Extract computation times (if they exist) and reset them to zero.

        This method looks for the attributes:
          - 'output_computation_time'
          - 'jacobian_product_computation_time'

        These attributes should have been used to count the total computation times
        for evaluating the function and its matrix jacobian product respectively

        Returns:
            A dictionary with keys:
              'output_computation_time' and
              'jacobian_product_computation_time' containing their values.
              If an attribute is not set, its value will be None.
        """
        times: Dict[str, Optional[float]] = {}
        if hasattr(self, "output_computation_time"):
            times["output_computation_time"] = self.output_computation_time
            self.output_computation_time = 0.0
        else:
            times["output_computation_time"] = None

        if hasattr(self, "jacobian_product_computation_time"):
            times["jacobian_product_computation_time"] = self.jacobian_product_computation_time
            self.jacobian_product_computation_time = 0.0
        else:
            times["jacobian_product_computation_time"] = None

        return times


def generate_full_Jacobian_data_for_computing_dimension_reduction(
    func_wrapper: FunctionAndDerivatives, N_samples: int
) -> Dict[str, np.ndarray]:
    """
    Generate full Jacobian data for dimension reduction.

    This is a thin wrapper around generate_reduced_training_data, which only
    requires an object that accepts NumPy arrays as input and returns NumPy arrays.

    Parameters:
        func_wrapper: An object implementing FunctionAndDerivatives.
        N_samples: The number of samples to generate.

    Returns:
        A dictionary containing training data and optional computation times.
    """
    return generate_reduced_training_data(func_wrapper, N_samples)


def generate_reduced_training_data(
    func_wrapper: FunctionAndDerivatives,
    N_samples: int,
    output_encoder: Optional[np.ndarray] = None,
    input_decoder: Optional[np.ndarray] = None,
    input_encoder: Optional[np.ndarray] = None,
) -> Dict[str, np.ndarray]:
    """
    Generate reduced training data using a function and its derivatives.

    This function repeatedly samples inputs and computes the corresponding
    function outputs and matrix–Jacobian products. It then applies optional
    encoding/decoding to reduce the dimensionality of the inputs, outputs, and
    Jacobian products. This function requires that the provided func_wrapper works
    entirely with NumPy arrays.

    Parameters:
        func_wrapper: An object implementing FunctionAndDerivatives that works with
            NumPy arrays.
        N_samples: The number of samples to generate.
        output_encoder: Optional; a NumPy array of shape
            (func_wrapper.output_dimension, reduced_out_dim) to encode outputs.
        input_decoder: Optional; a NumPy array of shape
            (func_wrapper.input_dimension, reduced_in_dim) to reduce the input
            dimension of the Jacobian product.
        input_encoder: Optional; a NumPy array of shape
            (func_wrapper.input_dimension, reduced_in_dim) to encode inputs.

    Returns:
        A dictionary with keys:
          'inputs': A NumPy array of shape (N_samples, D_in_encoded), where
              D_in_encoded equals reduced_in_dim if input_encoder is provided,
              otherwise func_wrapper.input_dimension.
          'outputs': A NumPy array of shape (N_samples, D_out_encoded), where
              D_out_encoded equals reduced_out_dim if output_encoder is provided,
              otherwise func_wrapper.output_dimension.
          'Jacobians': A NumPy array whose shape depends on the provided
              encoders/decoders:
                - (N_samples, reduced_out_dim, reduced_in_dim) if both
                  output_encoder and input_decoder are provided.
                - (N_samples, reduced_out_dim, func_wrapper.input_dimension) if only
                  output_encoder is provided.
                - (N_samples, func_wrapper.output_dimension, reduced_in_dim) if only
                  input_decoder is provided.
                - (N_samples, func_wrapper.output_dimension,
                  func_wrapper.input_dimension) if neither is provided.
          'output_computation_time': The time for computing outputs (if set).
          'Jacobian_computation_time': The time for computing the Jacobian product
              (if set).
    """
    encoded_input_dimension = input_encoder.shape[1] if input_encoder is not None else func_wrapper.input_dimension
    encoded_output_dimension = output_encoder.shape[1] if output_encoder is not None else func_wrapper.output_dimension
    print("Preparing for sampling...")
    encoded_inputs = np.empty((N_samples, encoded_input_dimension))
    encoded_outputs = np.empty((N_samples, encoded_output_dimension))
    encoded_jacobian_prod = np.empty((N_samples, encoded_output_dimension, encoded_input_dimension))

    func_wrapper.set_matrix_jacobian_prod(matrix=output_encoder)
    print("Sampling...")
    input_encoding_time = 0.0
    output_encoding_time = 0.0
    jacobian_encoding_time = 0.0
    for i in range(N_samples):
        input_sample = func_wrapper.sample_input()
        if input_encoder is not None:
            start = time.time()
            encoded_inputs[i] = input_sample @ input_encoder
            input_encoding_time += time.time() - start

        else:
            encoded_inputs[i] = input_sample
        output_i, matrix_jacobian_prod_i = func_wrapper.value_and_matrix_jacobian_prod(input_sample)
        if output_encoder is not None:
            start = time.time()
            encoded_outputs[i] = output_i @ output_encoder
            output_encoding_time += time.time() - start
        else:
            encoded_outputs[i] = output_i

        if input_decoder is not None:
            start = time.time()
            encoded_jacobian_prod[i] = matrix_jacobian_prod_i @ input_decoder
            jacobian_encoding_time += time.time() - start
        else:
            encoded_jacobian_prod[i] = matrix_jacobian_prod_i

    computation_times = func_wrapper.extract_and_clear_computation_times()
    input_key = "encoded_inputs" if input_encoder is not None else "inputs"
    output_key = "encoded_output" if output_encoder is not None else "outputs"
    jacobian_key = "encoded_Jacobians" if (input_decoder is not None or output_encoder is not None) else "Jacobians"

    results = {
        input_key: encoded_inputs,
        output_key: encoded_outputs,
        jacobian_key: encoded_jacobian_prod,
        "output_computation_time": computation_times.get("output_computation_time"),
        "Jacobian_computation_time": computation_times.get("jacobian_product_computation_time"),
    }
    if input_encoder is not None:
        results["input_encoding_time"] = input_encoding_time
    if output_encoder is not None:
        results["output_encoding_time"] = output_encoding_time
    if input_decoder is not None:
        results["Jacobian_encoding_time"] = jacobian_encoding_time
    return results


def encode_inputs(*, inputs: jnp.ndarray, encoder: jnp.ndarray) -> jnp.ndarray:
    """
    Encodes input data using the provided encoder matrix.

    Parameters:
      inputs: jax array, shape (n_samples, input_dim).
      encoder: jax array, shape (input_dim, reduced_dim).

    Returns:
      jax array, shape (n_samples, reduced_dim), with encoded inputs.
    """
    return jnp.einsum("nx,xr->nr", inputs, encoder)


def encode_outputs(*, outputs: jnp.ndarray, encoder: jnp.ndarray) -> jnp.ndarray:
    """
    Encodes output data using the provided encoder matrix.

    Parameters:
      outputs: jax array, shape (n_samples, output_dim).
      encoder: jax array, shape (output_dim, reduced_dim).

    Returns:
      jax array, shape (n_samples, reduced_dim), with encoded outputs.
    """
    return jnp.einsum("no,or->nr", outputs, encoder)


def encode_Jacobians(
    *,
    jacobians: jnp.ndarray,
    input_decoder: Optional[jnp.ndarray] = None,
    output_encoder: Optional[jnp.ndarray] = None,
    batched: bool = False,
    batch_size: int = 50,
) -> jnp.ndarray:
    """
    Reduces Jacobians using input decoder and/or output encoder matrices.

    Parameters:
      jacobians: jax array, shape (n_samples, output_dim, input_dim).
      input_decoder: Optional; jax array of shape
        (input_dim, reduced_in_dim).
      output_encoder: Optional; jax array of shape
        (output_dim, reduced_out_dim).
      batched: Process reduction in batches (default: False).
      batch_size: Batch size when batched is True (default: 50).

    Returns:
      jax array with reduced dimensions:
        - If both input_decoder and output_encoder provided:
            (n_samples, reduced_out_dim, reduced_in_dim)
        - If only output_encoder provided:
            (n_samples, reduced_out_dim, input_dim)
        - If only input_decoder provided:
            (n_samples, output_dim, reduced_in_dim)
        - If neither provided, returns original jacobians.
    """

    def reduce_batch(jacs: jnp.ndarray) -> jnp.ndarray:
        if output_encoder is not None and input_decoder is not None:
            return jnp.einsum("ol,nox,xr->nlr", output_encoder, jacs, input_decoder)
        elif output_encoder is not None:
            return jnp.einsum("ol,nox->nlx", output_encoder, jacs)
        elif input_decoder is not None:
            return jnp.einsum("nox,xr->nor", jacs, input_decoder)
        else:
            return jacs

    if batched:
        total_len = jacobians.shape[0]
        reduced_batches = []
        for start in range(0, total_len, batch_size):
            end = min(start + batch_size, total_len)
            batch = jax.device_put(jacobians[start:end])
            reduced_batches.append(reduce_batch(batch))
        return jnp.concatenate(reduced_batches, axis=0)
    else:
        return reduce_batch(jacobians)


def encode_input_output_Jacobian_data(
    *,
    inputs: jnp.ndarray,
    outputs: jnp.ndarray,
    jacobians: jnp.ndarray,
    input_encoder: Optional[jnp.ndarray] = None,
    output_encoder: Optional[jnp.ndarray] = None,
    input_decoder: Optional[jnp.ndarray] = None,
    batched: bool = False,
    batch_size: int = 50,
) -> Dict[str, jnp.ndarray]:
    """
    Encodes inputs, outputs, and reduces Jacobians using the provided matrices.

    Parameters:
      inputs: jax array, shape (n_samples, input_dim).
      outputs: jax array, shape (n_samples, output_dim).
      jacobians: jax array, shape (n_samples, output_dim, input_dim).
      input_encoder: Optional; for encoding inputs, shape
        (input_dim, reduced_dim).
      output_encoder: Optional; for encoding outputs, shape
        (output_dim, reduced_dim).
      input_decoder: Optional; for reducing jacobian input dim, shape
        (input_dim, reduced_in_dim).
      batched: Process jacobians in batches (default: False).
      batch_size: Batch size if batched is True.

    Returns:
      Dict with keys:
        "encoded_inputs": encoded inputs (or original if not provided).
        "encoded_outputs": encoded outputs (or original if not provided).
        "reduced_Jacobians": reduced jacobians (or original if not reduced).
    """
    inputs = jax.device_put(inputs) if inputs is not None else None
    outputs = jax.device_put(outputs) if outputs is not None else None
    input_decoder = jax.device_put(input_decoder) if input_decoder is not None else None
    input_encoder = jax.device_put(input_encoder) if input_encoder is not None else None
    output_encoder = jax.device_put(output_encoder) if output_encoder is not None else None

    start = time.time()
    encoded_inputs = encode_inputs(inputs=inputs, encoder=input_encoder) if input_encoder is not None else inputs
    input_encoding_time = time.time() - start

    start = time.time()
    encoded_outputs = encode_outputs(outputs=outputs, encoder=output_encoder) if output_encoder is not None else outputs
    output_encoding_time = time.time() - start

    start = time.time()
    encoded_Jacobians = encode_Jacobians(
        jacobians=jacobians,
        input_decoder=input_decoder,
        output_encoder=output_encoder,
        batched=batched,
        batch_size=batch_size,
    )
    jacobian_encoding_time = time.time() - start
    return {
        "encoded_inputs": encoded_inputs,
        "encoded_outputs": encoded_outputs,
        "encoded_Jacobians": encoded_Jacobians,
        "input_encoding_time": input_encoding_time,
        "output_encoding_time": output_encoding_time,
        "Jacobian_encoding_time": jacobian_encoding_time,
    }



import time
from typing import Any, Dict

import jax
import jax.numpy as jnp
from jax.scipy.linalg import solve_triangular
from randlax import (double_pass_randomized_eigh,
                     double_pass_randomized_gen_eigh)


@jax.jit
def __JCJtransp(J: jnp.ndarray, L: jnp.ndarray) -> jnp.ndarray:
    # Solve L Y = J.T for Y.
    Y = solve_triangular(L, J.T, lower=True)
    # Then, J C^-1 J.T = Y.T @ Y
    return Y.T @ Y


def average_JCJtranspose(J_samples: jnp.ndarray, prior_precision: jnp.ndarray) -> jnp.ndarray:
    L = jnp.linalg.cholesky(prior_precision)
    # Average over the batch dimension (axis=0)
    return jnp.mean(jax.vmap(lambda J: __JCJtransp(J, L))(J_samples), axis=0)


def information_theoretic_dimension_reduction(
    key: Any,
    J_samples: jnp.ndarray,
    noise_variance: float,
    max_input_dimension: int,
    max_output_dimension: int,
    prior_precision: jnp.ndarray,
    prior_covariance: jnp.ndarray = None,
):
    start = time.time()
    input_encodec_dict = estimate_input_active_subspace(
        key=key,
        J_samples=J_samples,
        noise_variance=noise_variance,
        prior_precision=prior_precision,
        subspace_rank=max_input_dimension,
    )
    input_encodec_dict["computation_time"] = time.time() - start

    start = time.time()
    output_encodec_dict = estimate_output_informative_subspace(
        key=key,
        J_samples=J_samples,
        noise_variance=noise_variance,
        subspace_rank=max_output_dimension,
        prior_precision=prior_precision,
        prior_covariance=prior_covariance,
    )
    output_encodec_dict["computation_time"] = time.time() - start
    return {"input": input_encodec_dict, "output": output_encodec_dict}


def estimate_input_active_subspace(
    key: Any,
    J_samples: jnp.ndarray,
    noise_variance: float,
    prior_precision: jnp.ndarray,
    subspace_rank: int,
) -> Dict[str, jnp.ndarray]:
    """
    Compute the rank-r active subspace encoder and decoder matrices.

    This function is applicable for finding the active subspace of the
    parameter under a Gaussian prior and additive Gaussian noise.
    It computes the matrix A as the mean of:
      J_samples[i]^T * (1/noise_variance) * J_samples[i]
    over all samples, and then uses a double-pass randomized generalized
    eigendecomposition to obtain the eigenvalues and eigenvectors. The
    decoder is the eigenvectors, while the encoder is given by
    prior_precision @ eigenvectors.

    Parameters:
      key: Random key for the randomized eigendecomposition.
      J_samples: 3D array of Jacobian samples with shape (N, a, b), where N
                is the number of samples.
      noise_variance: a positive scalar; noise precision is
                      1/noise_variance.
      prior_precision: 2D prior precision matrix.
      r: Target rank for the eigendecomposition.

    Returns:
      Dict with:
        "eigenvalues": 1D array of computed eigenvalues.
        "decoder": 2D array of computed eigenvectors.
        "encoder": 2D array computed as prior_precision @ eigenvectors.
    """

    A = jnp.einsum("iab,iac->bc", J_samples, J_samples / J_samples.shape[0]) / noise_variance
    computed_eigvals, computed_evecs = double_pass_randomized_gen_eigh(
        key,
        A,
        prior_precision,
        subspace_rank,
        subspace_rank + 15,
        power_iters=2,
        reorthog_iter=5,
    )
    return {
        "eigenvalues": computed_eigvals,
        "decoder": computed_evecs,
        "encoder": prior_precision @ computed_evecs,
    }


def estimate_output_informative_subspace(
    key: Any,
    J_samples: jnp.ndarray,
    noise_variance: float,
    subspace_rank: int,
    prior_covariance: jnp.ndarray = None,
    prior_precision: jnp.ndarray = None,
) -> Dict[str, jnp.ndarray]:
    """
    Compute the rank-r encoder and decoder for informative data.

    This function is used for computing the informative data subspace
    under a Gaussian prior and additive Gaussian noise. It forms the
    matrix A as the mean over samples:
      A = (1/N) * sum(J_samples[i] * (prior_covariance/noise_variance) *
                      J_samples[i]^T)
    A double-pass randomized eigendecomposition is applied to obtain the
    eigenvalues and eigenvectors. The decoder scales the eigenvectors by
    sqrt(noise_variance), while the encoder scales them by the inverse of
    sqrt(noise_variance).

    Parameters:
      key: Random key for the eigendecomposition.
      J_samples: 3D array of Jacobian samples with shape (N, a, b), where N
                is the number of samples.
      noise_variance: positive scalar.
      r: Target rank for the eigendecomposition.
      prior_covariance: 2D prior covariance matrix.
      prior_precision: 2D prior precision matrix. If prior covariance is not
        provided, the prior precision will be used to perform a solve

    Returns:
      Dict with:
        "eigenvalues": 1D array of computed eigenvalues.
        "decoder": 2D array (scaled eigenvectors).
        "encoder": 2D array (inversely scaled eigenvectors).
    """
    if prior_covariance is None:
        A = average_JCJtranspose(J_samples, prior_precision)
    else:
        A = jnp.einsum(
            "iab,bc,idc->ad",
            J_samples,
            prior_covariance / noise_variance,
            J_samples / J_samples.shape[0],
        )
    computed_eigvals, computed_evecs = double_pass_randomized_eigh(
        key, A, subspace_rank, p=subspace_rank + 15, power_iters=2
    )
    noise_sigma = jnp.sqrt(noise_variance)
    return {
        "eigenvalues": computed_eigvals,
        "decoder": noise_sigma * computed_evecs,
        "encoder": computed_evecs / noise_sigma,
    }
